#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>

#include "rndpassw.h"

int main(int argc, char **argv) {
	int i, ch, passlen, passcnt, dictsz, rndsz, passsz;
	u_int16_t *rndbuf, *rndoff;
	char *dictbuf, *passbuf;
	rndopt_t rndopt = { 0 };

	while((ch = getopt(argc, argv, "dusptvh")) != -1) {
		switch (ch) {
			case 'd':
				rndopt.digit = 1;
				break;

			case 'u':
				rndopt.ualpha = 1;
				break;

			case 's':
				rndopt.salpha = 1;
				break;

			case 'p':
				rndopt.print = 1;
				break;

			case 't':
				rndopt.digit = 1;
				rndopt.ualpha = 1;
				rndopt.salpha = 1;
				rndopt.print = 1;
				break;

			case 'h':
				usage();
				return(EXIT_SUCCESS);

			case 'v':
				version();
				return(EXIT_SUCCESS);
		}
	}

	argc -= optind;
	argv += optind;

	if(argv[0] != NULL) {
		passlen = atoi(argv[0]);
	} else {
		passlen = PASSLEN;
	}

	if((argv[1] != NULL) && (argv[0] != NULL)) {
		passcnt = atoi(argv[1]);
	} else {
		passcnt = PASSCNT;
	}
	
	if(passlen > MAXPASSLEN) {
		printf("Error: max passwords length is %d\n", MAXPASSLEN);
		return(EXIT_FAILURE);
	}

	if(passcnt > MAXPASSCNT) {
		printf("Error: max passwords count is %d\n", MAXPASSCNT);
		return(EXIT_FAILURE);
	}

	if((rndopt.digit + rndopt.ualpha + rndopt.salpha +
		rndopt.print) == 0) {
		rndopt.digit = 1;
		rndopt.ualpha = 1;
		rndopt.salpha = 1;
	}

	dictsz = (rndopt.digit * SZ_DIGIT);
	dictsz += (rndopt.ualpha * SZ_UALPHA);
	dictsz += (rndopt.salpha * SZ_SALPHA);
	dictsz += (rndopt.print * SZ_PRINT);

	passsz = passlen * passcnt * sizeof(u_int16_t);
	rndsz = passsz + (dictsz * sizeof(u_int16_t));

	rndbuf = malloc(rndsz);

	if(rndbuf == NULL) {
		printf("Error: could not allocate memory\n");
		return(EXIT_FAILURE);
	}

	dictbuf = malloc(dictsz);

	if(dictbuf == NULL) {
		printf("Error: could not allocate memory\n");
		return(EXIT_FAILURE);
	}

	passbuf = calloc(passlen + 1, 1);

	if(passbuf == NULL) {
		printf("Error: could not allocate memory\n");
		return(EXIT_FAILURE);
	}

	rndread(rndbuf, rndsz);

	if(rndbuf == NULL) {
		printf("Error: could read \"/dev/urandom\"\n");
		return(EXIT_FAILURE);
	}

	rndoff = rndbuf;

	mixdict(dictbuf, rndopt);

	sortdict(dictbuf, dictsz, rndoff);

	for(i = 0; i < passcnt; i++) {
		genpass(passbuf, passlen, dictbuf, dictsz, rndoff);
		printf("%s\n", passbuf);
		rndoff+=passlen;
	}

	free(rndbuf);
	free(passbuf);
	free(dictbuf);

	return(EXIT_SUCCESS);
}

void usage() {
	printf(	"Usage: rndpassw [-duspth] <passlen> <passcount>\n"
		" Options:\n"
		"  -d\t - numeric characters\n"
		"  -u\t - upper alpha characters\n"
		"  -s\t - small alpha characters\n"
		"  -p\t - printable charaters\n"
		"  -t\t - true secure (include all groups)\n"
		"  -v\t - print version\n"
		"  -h\t - print this help\n");
}

void version() {
	printf("%d.%d\n", VERMAJOR, VERMINOR);
}

void mixdict(char *dictbuf, rndopt_t rndopt) {
	int offset = 0;

	if(rndopt.digit == 1) {
		memcpy(dictbuf, dict_digit, SZ_DIGIT);
		offset += SZ_DIGIT;
	}

	if(rndopt.ualpha == 1) {
		memcpy(dictbuf + offset, dict_ualpha, SZ_UALPHA);
		offset += SZ_UALPHA;
	}

	if(rndopt.salpha == 1) {
		memcpy(dictbuf + offset, dict_salpha, SZ_SALPHA);
		offset += SZ_SALPHA;
	}

	if(rndopt.print == 1) {
		memcpy(dictbuf + offset, dict_print, SZ_PRINT);
	}
}

void rndread(u_int16_t *rndbuf, ssize_t rndsz) {
	int fd;
	ssize_t len;

	fd = open("/dev/urandom", O_RDONLY);

        if(fd == -1) {
		rndbuf = NULL;
                return;
        }

        len = read(fd, rndbuf, rndsz);

        if(len != rndsz) {
		rndbuf = NULL;
                return;
        }

        close(fd);		
}

void sortdict(char *dictbuf, int dictsz, u_int16_t *rndbuf) {
	int i, swap;
	char tmp;

	for(i = 0; i < dictsz; i++) {
		swap = rndbuf[i] % dictsz;
		tmp = dictbuf[i];
		dictbuf[i] = dictbuf[swap];
		dictbuf[swap] = tmp;
	}

	rndbuf+=i;
}

void genpass(char *passbuf, int passlen, char *dictbuf, int dictsz, u_int16_t *rndbuf) {
	int i, z;

	for(i = 0; i < passlen; i++) {
		z = rndbuf[i] % dictsz;
		passbuf[i] = dictbuf[z];
	}
}

